@@ -4,12 +4,12 @@ Name Server and helper functions.
 Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
 """
 
-from __future__ import with_statement
 import warnings
 import re
 import logging
 import socket
 import sys
+import time
 from Pyro4.threadutil import RLock, Thread
 from Pyro4.errors import NamingError, PyroError, ProtocolError
 from Pyro4 import core, socketutil
@@ -61,6 +61,7 @@ class MemoryStorage(dict):
         pass
 
 
+@Pyro4.expose
 class NameServer(object):
     """
     Pyro name server. Provides a simple flat name space to map logical object names to Pyro URIs.
@@ -84,7 +85,10 @@ class NameServer(object):
         try:
             uri, metadata = self.storage[name]
             uri = core.URI(uri)
-            return (uri, metadata or frozenset()) if return_metadata else uri
+            if return_metadata:
+                metadata = list(metadata) if metadata else []
+                return uri, metadata
+            return uri
         except KeyError:
             raise NamingError("unknown name: " + name)
 
@@ -150,22 +154,32 @@ class NameServer(object):
         """Retrieve the registered items as a dictionary name-to-URI. The URIs
         in the resulting dict are strings, not URI objects.
         You can filter by prefix or by regex or by metadata subset (separately)"""
+        def fix_set(result):
+            # for python 2 compatibility we cannot send sets to the default (serpent) serializer.
+            # that's why we will convert them to lists here.
+            if return_metadata:
+                fixed = {}
+                for name, data in result.items():
+                    fixed[name] = (data[0], list(data[1]))
+                return fixed
+            return result
+
         if sum(1 for x in [prefix, regex, metadata_all, metadata_any] if x is not None) > 1:
             raise ValueError("you can only filter on one thing at a time")
         with self.lock:
             if prefix:
                 result = self.storage.optimized_prefix_list(prefix, return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 result = {}
                 for name in self.storage:
                     if name.startswith(prefix):
                         result[name] = self.storage[name] if return_metadata else self.storage[name][0]
-                return result
+                return fix_set(result)
             elif regex:
                 result = self.storage.optimized_regex_list(regex, return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 result = {}
                 try:
                     regex = re.compile(regex)
@@ -176,7 +190,7 @@ class NameServer(object):
                     for name in self.storage:
                         if regex.match(name):
                             result[name] = self.storage[name] if return_metadata else self.storage[name][0]
-                    return result
+                    return fix_set(result)
             elif metadata_all:
                 # return the entries which have all of the given metadata as (a subset of) their metadata
                 if isinstance(metadata_all, basestring):
@@ -184,13 +198,13 @@ class NameServer(object):
                 metadata_all and iter(metadata_all)   # validate that metadata is iterable
                 result = self.storage.optimized_metadata_search(metadata_all=metadata_all, return_metadata=return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 metadata_all = frozenset(metadata_all)
                 result = {}
                 for name, (uri, meta) in self.storage.everything(return_metadata=True).items():
                     if metadata_all.issubset(meta):
                         result[name] = (uri, meta) if return_metadata else uri
-                return result
+                return fix_set(result)
             elif metadata_any:
                 # return the entries which have any of the given metadata as part of their metadata
                 if isinstance(metadata_any, basestring):
@@ -198,16 +212,16 @@ class NameServer(object):
                 metadata_any and iter(metadata_any)   # validate that metadata is iterable
                 result = self.storage.optimized_metadata_search(metadata_any=metadata_any, return_metadata=return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 metadata_any = frozenset(metadata_any)
                 result = {}
                 for name, (uri, meta) in self.storage.everything(return_metadata=True).items():
                     if metadata_any & meta:
                         result[name] = (uri, meta) if return_metadata else uri
-                return result
+                return fix_set(result)
             else:
                 # just return (a copy of) everything
-                return self.storage.everything(return_metadata)
+                return fix_set(self.storage.everything(return_metadata))
 
     def ping(self):
         """A simple test method to check if the name server is running correctly."""
@@ -252,6 +266,15 @@ class NameServerDaemon(core.Daemon):
         self.register(self.nameserver, Pyro4.constants.NAMESERVER_NAME)
         metadata = {"class:Pyro4.naming.NameServer"}
         self.nameserver.register(Pyro4.constants.NAMESERVER_NAME, self.uriFor(self.nameserver), metadata=metadata)
+        if Pyro4.config.NS_AUTOCLEAN > 0:
+            if not AutoCleaner.override_autoclean_min and Pyro4.config.NS_AUTOCLEAN < AutoCleaner.min_autoclean_value:
+                raise ValueError("NS_AUTOCLEAN cannot be smaller than " + str(AutoCleaner.min_autoclean_value))
+            log.debug("autoclean enabled")
+            self.cleaner_thread = AutoCleaner(self.nameserver)
+            self.cleaner_thread.start()
+        else:
+            log.debug("autoclean not enabled")
+            self.cleaner_thread = None
         log.info("nameserver daemon created")
 
     def close(self):
@@ -259,6 +282,10 @@ class NameServerDaemon(core.Daemon):
         if self.nameserver is not None:
             self.nameserver.storage.close()
             self.nameserver = None
+        if self.cleaner_thread:
+            self.cleaner_thread.stop = True
+            self.cleaner_thread.join()
+            self.cleaner_thread = None
 
     def __enter__(self):
         if not self.nameserver:
@@ -269,6 +296,10 @@ class NameServerDaemon(core.Daemon):
         if self.nameserver is not None:
             self.nameserver.storage.close()
         self.nameserver = None
+        if self.cleaner_thread:
+            self.cleaner_thread.stop = True
+            self.cleaner_thread.join()
+            self.cleaner_thread = None
         return super(NameServerDaemon, self).__exit__(exc_type, exc_value, traceback)
 
     def handleRequest(self, conn):
@@ -282,10 +313,71 @@ class NameServerDaemon(core.Daemon):
             raise
 
 
+class AutoCleaner(Thread):
+    """
+    Takes care of checking every registration in the name server.
+    If it cannot be contacted anymore, it will be removed after ~20 seconds.
+    """
+    min_autoclean_value = 3
+    max_unreachable_time = 20.0
+    loop_delay = 2.0
+    override_autoclean_min = False   # only for unit test purposes
+
+    def __init__(self, nameserver):
+        assert Pyro4.config.NS_AUTOCLEAN > 0
+        if not self.override_autoclean_min and Pyro4.config.NS_AUTOCLEAN < self.min_autoclean_value:
+            raise ValueError("NS_AUTOCLEAN cannot be smaller than "+str(self.min_autoclean_value))
+        super(AutoCleaner, self).__init__()
+        self.nameserver = nameserver
+        self.stop = False
+        self.daemon = True
+        self.last_cleaned = time.time()
+        self.unreachable = {}   # name->since when
+
+    def run(self):
+        while not self.stop:
+            time.sleep(self.loop_delay)
+            time_since_last_autoclean = time.time() - self.last_cleaned
+            if time_since_last_autoclean < Pyro4.config.NS_AUTOCLEAN:
+                continue
+            for name, uri in self.nameserver.list().items():
+                if name in (Pyro4.constants.DAEMON_NAME, Pyro4.constants.NAMESERVER_NAME):
+                    continue
+                try:
+                    uri_obj = Pyro4.URI(uri)
+                    timeout = Pyro4.config.COMMTIMEOUT or 5
+                    sock = socketutil.createSocket(connect=(uri_obj.host, uri_obj.port), timeout=timeout)
+                    sock.close()
+                    # if we get here, the listed server is still answering on its port
+                    if name in self.unreachable:
+                        del self.unreachable[name]
+                except socket.error:
+                    if name not in self.unreachable:
+                        self.unreachable[name] = time.time()
+                    if time.time() - self.unreachable[name] >= self.max_unreachable_time:
+                        log.info("autoclean: unregistering %s; cannot connect uri %s for %d sec", name, uri, self.max_unreachable_time)
+                        self.nameserver.remove(name)
+                        del self.unreachable[name]
+                        continue
+            self.last_cleaned = time.time()
+            if self.unreachable:
+                log.debug("autoclean: %d/%d names currently unreachable", len(self.unreachable), self.nameserver.count())
+
+
 class BroadcastServer(object):
-    REQUEST_NSURI = "GET_NSURI" if sys.platform == "cli" else b"GET_NSURI"
+    REQUEST_NSURI = b"GET_NSURI"
+
+    class TransportServerAdapter(object):
+        # this adapter is used to be able to pass the BroadcastServer to Daemon.combine() to integrate the event loops.
+        def __init__(self, bcserver):
+            self.sockets = [bcserver]
+
+        def events(self, eventobjects):
+            for bc in eventobjects:
+                bc.processRequest()
 
     def __init__(self, nsUri, bchost=None, bcport=None):
+        self.transportServer = self.TransportServerAdapter(self)
         self.nsUri = nsUri
         if bcport is None:
             bcport = Pyro4.config.NS_BCPORT
@@ -327,6 +419,7 @@ class BroadcastServer(object):
         thread.setDaemon(True)
         thread.start()
         log.debug("broadcast server loop running in own thread")
+        return thread
 
     def __requestLoop(self):
         while self.running:
@@ -351,7 +444,7 @@ class BroadcastServer(object):
         except socket.error:
             pass
         except SystemError:
-            if sys.platform == 'cli' and not self.running:
+            if sys.platform == "cli" and not self.running:
                 # ironpython throws these systemerrors when shutting down... we can ignore them.
                 pass
             else:
@@ -429,19 +522,27 @@ def locateNS(host=None, port=None, broadcast=True, hmac_key=None):
     if host is None:
         # first try localhost if we have a good chance of finding it there
         if Pyro4.config.NS_HOST in ("localhost", "::1") or Pyro4.config.NS_HOST.startswith("127."):
-            host = Pyro4.config.NS_HOST
-            if ":" in host:  # ipv6
-                host = "[%s]" % host
-            uristring = "PYRO:%s@%s:%d" % (Pyro4.constants.NAMESERVER_NAME, host, port or Pyro4.config.NS_PORT)
-            log.debug("locating the NS: %s", uristring)
-            proxy = core.Proxy(uristring)
-            proxy._pyroHmacKey = hmac_key
-            try:
-                proxy.ping()
-                log.debug("located NS")
-                return proxy
-            except PyroError:
-                pass
+            if ":" in Pyro4.config.NS_HOST:  # ipv6
+                hosts = ["[%s]" % Pyro4.config.NS_HOST]
+            else:
+                # Some systems (Debian Linux) have 127.0.1.1 in the hosts file assigned to the hostname,
+                # try this too for convenience sake (only if it's actually used as a valid ip address)
+                try:
+                    socket.gethostbyaddr("127.0.1.1")
+                    hosts = [Pyro4.config.NS_HOST] if Pyro4.config.NS_HOST == "127.0.1.1" else [Pyro4.config.NS_HOST, "127.0.1.1"]
+                except socket.error:
+                    hosts = [Pyro4.config.NS_HOST]
+            for host in hosts:
+                uristring = "PYRO:%s@%s:%d" % (Pyro4.constants.NAMESERVER_NAME, host, port or Pyro4.config.NS_PORT)
+                log.debug("locating the NS: %s", uristring)
+                proxy = core.Proxy(uristring)
+                proxy._pyroHmacKey = hmac_key
+                try:
+                    proxy._pyroBind()
+                    log.debug("located NS")
+                    return proxy
+                except PyroError:
+                    pass
         if broadcast:
             # broadcast lookup
             if not port:
@@ -495,7 +596,7 @@ def locateNS(host=None, port=None, broadcast=True, hmac_key=None):
     proxy = core.Proxy(uri)
     proxy._pyroHmacKey = hmac_key
     try:
-        proxy.ping()
+        proxy._pyroBind()
         log.debug("located NS")
         return proxy
     except PyroError as x:
